/**
 * Ext GWT 3.0.0-beta4 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.desktop.client.persistence;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.google.gwt.core.client.GWT;
import com.google.web.bindery.autobean.shared.AutoBean;
import com.google.web.bindery.autobean.shared.AutoBeanCodex;
import com.google.web.bindery.autobean.shared.AutoBeanFactory;
import com.google.web.bindery.autobean.shared.AutoBeanFactory.Category;
import com.google.web.bindery.autobean.shared.AutoBeanUtils;
import com.sencha.gxt.data.shared.ModelKeyProvider;
import com.sencha.gxt.data.shared.TreeStore;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent.StoreUpdateHandler;

/**
 * Models stored in a PersistentTreeStore must have been created using an
 * AutoBeanFactory.
 * 
 * @param <M> the model type
 */
public class PersistentTreeStore<M> extends TreeStore<M> {

  public interface StorageItem {
    List<String> getChildIds();

    String getData();

    String getId();

    String getParentId();

    void setChildIds(List<String> children);

    void setData(String data);

    void setId(String id);

    void setParentId(String parentId);
  }

  @Category(AutoBeanToString.class)
  public interface StorageItemFactory extends AutoBeanFactory {
    AutoBean<StorageItem> storageItem();
  }

  public class PersistentTreeModel extends TreeModel {

    private StorageItem storageItem;

    // called for root or when a new node is being added
    public PersistentTreeModel(M data) {
      super(data);
      if (isRoot()) {
        root = this;
      } else {
        initializeStorageItem(allocateId(PT_PREFIX));
      }
    }

    // called when restoring an existing node
    public PersistentTreeModel(M data, StorageItem storageItem) {
      super(data);
      this.storageItem = storageItem;
    }

    @Override
    public void addChild(int index, TreeModel child) {
      addChild(index, child, true);
    }

    public void addChild(int index, TreeModel child, boolean isAddLinkage) {
      super.addChild(index, child);
      if (isAddLinkage) {
        addParentChildLinkage(child);
      }
    }

    @Override
    public void addChildren(int index, List<TreeModel> children) {
      super.addChildren(index, children);
      for (TreeModel child : children) {
        addParentChildLinkage(child);
      }
    }

    @Override
    public void clear() {
      super.clear();
      clearChildren(storageItem);
      storageItem.getChildIds().clear();
    }

    @Override
    public void remove(TreeModel childTreeModel) {
      remove(childTreeModel, true);
    }

    @Override
    public String toString() {
      return storageItem == null ? "uninitialized" : storageItem.toString();
    }

    private void addParentChildLinkage(TreeModel childTreeModel) {
      @SuppressWarnings("unchecked")
      PersistentTreeModel persistentChildTreeModel = (PersistentTreeModel) childTreeModel;
      storageItem.getChildIds().add(persistentChildTreeModel.storageItem.getId());
      persistentChildTreeModel.storageItem.setParentId(storageItem.getId());
      persistentChildTreeModel.setData(childTreeModel.getData());
      writeStorageItem();
      persistentChildTreeModel.writeStorageItem();
    }

    private void clearChildren(StorageItem parentStorageItem) {
      for (String id : parentStorageItem.getChildIds()) {
        StorageItem childStorageItem = readItem(id);
        clearChildren(childStorageItem);
      }
      backingStore.removeItem(parentStorageItem.getId());
    }

    private StorageItem decode(String encodedStorageItem) {
      AutoBean<StorageItem> storageItemBean = AutoBeanCodex.decode(storageItemFactory, StorageItem.class,
          encodedStorageItem);
      return storageItemBean.as();
    }

    private M getData(StorageItem storageItem) {
      String encodedData = storageItem.getData();
      AutoBean<M> dataAutoBean = AutoBeanCodex.decode(dataFactory, dataClass, encodedData);
      return dataAutoBean.as();
    }

    private void initializeRoot() {
      String encodedStorageItem = backingStore.getItem(PT_PREFIX + ROOT_ID);
      if (encodedStorageItem == null) {
        initializeStorageItem(allocateId(PT_PREFIX));
      } else {
        restoreRootStorageItem(encodedStorageItem);
      }
    }

    private void initializeStorageItem(String id) {
      AutoBean<StorageItem> storageItemBean = storageItemFactory.create(StorageItem.class);
      storageItem = storageItemBean.as();
      storageItem.setId(id);
      storageItem.setChildIds(new LinkedList<String>());
      writeStorageItem();
    }

    private StorageItem readItem(String id) {
      String encodedStorageItem = backingStore.getItem(id);
      StorageItem storageItem = decode(encodedStorageItem);
      return storageItem;
    }

    private void remove(TreeModel childTreeModel, boolean isDeep) {
      super.remove(childTreeModel);
      @SuppressWarnings("unchecked")
      PersistentTreeModel persistentChildTreeModel = (PersistentTreeModel) childTreeModel;
      storageItem.getChildIds().remove(persistentChildTreeModel.storageItem.getId());
      if (isDeep) {
        clearChildren(persistentChildTreeModel.storageItem);
      }
      writeStorageItem();
    }

    private void restoreChildren(PersistentTreeModel parentTreeModel) {
      int index = 0;
      for (String childId : parentTreeModel.storageItem.getChildIds()) {
        String encodedChildStorageItem = backingStore.getItem(childId);
        if (encodedChildStorageItem == null) {
          throw new IllegalStateException("Missing child, parent=" + parentTreeModel.storageItem.getId() + ", child="
              + childId);
        } else {
          AutoBean<StorageItem> childStorageItemBean = AutoBeanCodex.decode(storageItemFactory, StorageItem.class,
              encodedChildStorageItem);
          StorageItem childStorageItem = childStorageItemBean.as();
          M data = getData(childStorageItem);
          PersistentTreeModel childTreeModel = new PersistentTreeModel(data, childStorageItem);
          getModelMap().put(getKeyProvider().getKey(data), childTreeModel);
          parentTreeModel.addChild(index++, childTreeModel, false);
          parentTreeModel.restoreChildren(childTreeModel);
        }
      }
    }

    private void restoreRootStorageItem(String encodedStorageItem) {
      storageItem = decode(encodedStorageItem);
      restoreChildren(this);
    }

    private void setData(M data) {
      AutoBean<M> dataAutoBean = AutoBeanUtils.getAutoBean(data);
      String encodedData = AutoBeanCodex.encode(dataAutoBean).getPayload();
      storageItem.setData(encodedData);
    }

    private void writeStorageItem() {
      AutoBean<StorageItem> storageItemBean = AutoBeanUtils.getAutoBean(storageItem);
      String encodedStorageItem = AutoBeanCodex.encode(storageItemBean).getPayload();
      String id = storageItem.getId();
      backingStore.setItem(id, encodedStorageItem);
    }

  }

  private static final String ROOT_ID = "0";
  private static final String ID_KEY = "id";
  private static final String PT_PREFIX = "p";

  private BackingStore backingStore;
  private AutoBeanFactory dataFactory;
  private Class<M> dataClass;
  private AutoBeanFactory storageItemFactory;
  private PersistentTreeModel root;

  public PersistentTreeStore(ModelKeyProvider<? super M> keyProvider, BackingStore backingStore,
      AutoBeanFactory dataFactory, Class<M> dataClass) {
    super(keyProvider);
    this.backingStore = backingStore;
    this.dataFactory = dataFactory;
    this.dataClass = dataClass;
    this.storageItemFactory = GWT.create(StorageItemFactory.class);
    root.initializeRoot(); // created by superclass before we're initialized
    addStoreUpdateHandler(new StoreUpdateHandler<M>() {
      // Handles both Store.update() and Record.addChange() row editing
      // Must not do anything that causes a recursive update
      @Override
      public void onUpdate(StoreUpdateEvent<M> event) {
        for (M item : event.getItems()) {
          persist(item);
        }
      }
    });
  }

  @SuppressWarnings("unchecked")
  public void move(List<?> childTreeModels, M newParent, int newIndex) {
    PersistentTreeModel newParentTreeModel;
    if (newParent == null) {
      newParent = root.getData();
      newParentTreeModel = root;
    } else {
      newParentTreeModel = (PersistentTreeModel) getWrapper(newParent);
    }
    // TODO: Factor out common code shared with TreeStore.remove
    for (PersistentTreeModel childTreeModel : (List<PersistentTreeModel>) childTreeModels) {
      PersistentTreeModel oldParentTreeModel = (PersistentTreeModel) childTreeModel.getParent();
      int oldIndex = oldParentTreeModel.getChildren().indexOf(childTreeModel);
      if (newParentTreeModel == oldParentTreeModel && oldIndex < newIndex) {
        newIndex--; // adjust for child being removed above insert point
      }
      oldParentTreeModel.remove(childTreeModel, false);
      M model = childTreeModel.getData();
      M parent = oldParentTreeModel.getData();
      List<M> children = getAllChildren(model);
      fireEvent(new PersistentTreeStoreRemoveEvent<M>(oldIndex, model, parent, children, true));
    }
    // TODO: Factor out common code shared with TreeStore.insert
    int initialCount = newParentTreeModel.getChildren().size();
    newParentTreeModel.addChildren(newIndex, (List<TreeModel>) childTreeModels);
    if (initialCount != newParentTreeModel.getChildren().size()) {
      List<M> addedChildren = new ArrayList<M>();
      List<TreeModel> currentChildren = newParentTreeModel.getChildren();
      int i = newIndex;
      for (M child : unwrap((List<TreeModel>) childTreeModels)) {
        if (currentChildren.get(i).getData() == child) {
          addedChildren.add(child);
          i++;
        }
      }
      if (addedChildren.size() != 0) {
        fireEvent(new PersistentTreeStoreAddEvent<M>(newIndex, addedChildren, true));
      }
    }
  }

  protected String allocateId(String idPrefix) {
    String thisId = backingStore.getItem(ID_KEY);
    if (thisId == null) {
      thisId = ROOT_ID;
    }
    String nextId = Integer.toString(Integer.parseInt(thisId) + 1);
    backingStore.setItem(ID_KEY, nextId);
    return idPrefix + thisId;
  }

  @Override
  protected TreeModel createTreeModel(M data) {
    return new PersistentTreeModel(data);
  }

  protected Class<M> getDataClass() {
    return dataClass;
  }

  protected AutoBeanFactory getDataFactory() {
    return dataFactory;
  }

  protected void persist(M data) {
    TreeStore<M>.TreeModel wrapper = getWrapper(data);
    if (wrapper == null) {
      throw new IllegalArgumentException("Cannot find wrapper, data=" + data);
    } else {
      @SuppressWarnings("unchecked")
      PersistentTreeModel persistentTreeModel = (PersistentTreeModel) wrapper;
      persistentTreeModel.setData(wrapper.getData());
      persistentTreeModel.writeStorageItem();
    }
  }

}
